@@ -1,22 +1,76 @@
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveGeneric #-}
+
 -- | WARNING: This module should be considered private! If you find yourself
 -- wanting to import something from this module, please open an issue to get
 -- that thing exported from "Salve".
-module Salve.Internal where
+module Salve.Internal
+  ( Version(..)
+  , PreRelease(..)
+  , Build(..)
+  , Constraint(..)
+  , makeVersion
+  , initialVersion
+  , parseVersion
+  , parsePreRelease
+  , parseBuild
+  , parseConstraint
+  , unsafeParseVersion
+  , unsafeParsePreRelease
+  , unsafeParseBuild
+  , unsafeParseConstraint
+  , renderVersion
+  , renderPreRelease
+  , renderBuild
+  , renderConstraint
+  , isUnstable
+  , isStable
+  , fromBaseVersion
+  , toBaseVersion
+  , bumpMajor
+  , bumpMinor
+  , bumpPatch
+  , satisfiesConstraint
+  , majorLens
+  , minorLens
+  , patchLens
+  , preReleasesLens
+  , buildsLens
+  , Operator(..)
+  , Wildcard(..)
+  , constraintLT
+  , constraintLE
+  , constraintEQ
+  , constraintGE
+  , constraintGT
+  , constraintAnd
+  , constraintOr
+  , constraintHyphen
+  , constraintTilde
+  , constraintCaret
+  ) where
 
 import qualified Control.Monad as Monad
 import qualified Data.Char as Char
+import qualified Data.Data as Data
 import qualified Data.List as List
 import qualified Data.Maybe as Maybe
 import qualified Data.Monoid as Monoid
 import qualified Data.Ord as Ord
 import qualified Data.Version as Version
 import qualified Data.Word as Word
+import qualified GHC.Generics as Generics
 import qualified Text.ParserCombinators.ReadP as ReadP
 
 -- $setup
--- >>> import Control.Applicative
--- >>> import Lens.Micro
--- >>> let view l x = x ^. l
+--
+-- >>> import Control.Applicative (Const(..))
+-- >>> let view lens record = getConst (lens Const record)
+--
+-- >>> import Data.Functor.Identity (Identity(..))
+-- >>> let set lens field record = runIdentity (lens (const (Identity field)) record)
+--
+-- >>> import Control.Applicative ((<$>), (<*>))
 
 -- * Public
 
@@ -35,7 +89,7 @@ data Version = Version
   , versionPatch :: Word.Word64
   , versionPreReleases :: [PreRelease]
   , versionBuilds :: [Build]
-  } deriving (Eq, Show)
+  } deriving (Data.Data, Eq, Generics.Generic, Read, Show)
 
 -- | In general, versions compare in the way that you would expect. First the
 -- major version numbers are compared, then the minors, then the patches.
@@ -98,7 +152,7 @@ instance Ord Version where
 data PreRelease
   = PreReleaseNumeric Word.Word64
   | PreReleaseTextual String
-  deriving (Eq, Show)
+  deriving (Data.Data, Eq, Generics.Generic, Read, Show)
 
 -- | Numeric pre-releases are always less than textual pre-releases.
 --
@@ -130,7 +184,8 @@ instance Ord PreRelease where
 -- 3. Builds cannot be compared. That is, they do not have an 'Ord' instance.
 --
 -- Use 'parseBuild' to create builds.
-newtype Build = Build String deriving (Eq, Show)
+newtype Build = Build String
+  deriving (Data.Data, Eq, Generics.Generic, Read, Show)
 
 -- | Constrains allowable version numbers.
 --
@@ -142,7 +197,7 @@ data Constraint
   | ConstraintWildcard Wildcard
   | ConstraintAnd Constraint Constraint
   | ConstraintOr Constraint Constraint
-  deriving (Eq, Show)
+  deriving (Data.Data, Eq, Generics.Generic, Ord, Read, Show)
 
 -- | Makes a new version number.
 --
@@ -178,12 +233,18 @@ initialVersion = makeVersion 0 0 0 [] []
 -- >>> parseVersion "wrong"
 -- Nothing
 --
--- Whitespace is not allowed and will cause the parser to fail.
+-- Whitespace is allowed.
 --
 -- >>> parseVersion " 1.2.3 "
--- Nothing
+-- Just (Version {versionMajor = 1, versionMinor = 2, versionPatch = 3, versionPreReleases = [], versionBuilds = []})
 parseVersion :: String -> Maybe Version
-parseVersion s = parse versionP s
+parseVersion s = parse
+  (do
+    ReadP.skipSpaces
+    version <- versionP
+    ReadP.skipSpaces
+    return version)
+  s
 
 -- | Attempts to parse a pre-release.
 --
@@ -586,13 +647,13 @@ data Operator
   | OperatorGT
   | OperatorTilde
   | OperatorCaret
-  deriving (Eq, Show)
+  deriving (Data.Data, Eq, Generics.Generic, Ord, Read, Show)
 
 data Wildcard
   = WildcardMajor
   | WildcardMinor Word.Word64
   | WildcardPatch Word.Word64 Word.Word64
-  deriving (Eq, Show)
+  deriving (Data.Data, Eq, Generics.Generic, Ord, Read, Show)
 
 -- ** Constructors
 
@@ -746,7 +807,7 @@ nonZeroP = do
   ys <- ReadP.munch Char.isDigit
   case toWord64 (stringToIntegral (x : ys)) of
     Nothing -> ReadP.pfail
-    Just n -> pure n
+    Just n -> return n
 
 constraintsP :: ReadP.ReadP Constraint
 constraintsP = do
@@ -899,7 +960,7 @@ data SimpleConstraint
   | SCGT Version
   | SCAnd SimpleConstraint SimpleConstraint
   | SCOr SimpleConstraint SimpleConstraint
-  deriving (Eq, Show)
+  deriving (Data.Data, Eq, Generics.Generic, Ord, Read, Show)
 
 mkV :: Word.Word64 -> Word.Word64 -> Word.Word64 -> Version
 mkV m n p = makeVersion m n p [] []
